#include <QProcess>

#include "about.h"
#include "macros.h"
#include "options.h"
#include "mameopt.h"
#include "qmc2main.h"
#include "gamelist.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern Gamelist *qmc2Gamelist;
extern MameOptions *qmc2GlobalMAMEOptions;

About::About(QWidget *parent)
  : QDialog(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: About::About(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  setupUi(this);
}

About::~About()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: About::~About()");
#endif

}

void About::showEvent(QShowEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: About::showEvent(QShowEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  QString titleString =
          "<p><font size=\"+1\"><b>QMC2 - M.A.M.E. Catalog / Launcher II</b></font><br>" + tr("Qt 4 based UNIX multi-emulator frontend") + "<br>" +
          tr("Version ") + QString::number(MAJOR) + "." + QString::number(MINOR) +
#ifdef BETA
          ".b" + QString::number(BETA) +
#endif
          ", " + tr("built for") + 
#ifdef SDLMAME
          " SDLMAME" +
#else
          " ???" +
#endif
          "<br>" + tr("Copyright") + " &copy; 2006 - 2008 R. Reucher, " + tr("Germany") + "</p>";
  labelTitle->setText(titleString);

  QString projectInfoString =
          "<p><b>" + tr("Project homepage:") + "</b><br>http://www.mameworld.net/mamecat</p>" +
          "<p><b>" + tr("Development site:") + "</b><br>http://sourceforge.net/projects/qmc2</p>" +
          "<p><b>" + tr("QMC2 development mailing list:") + "</b><br>qmc2-devel@lists.sourceforge.net (" + tr("subscribers only") +")</p>" +
          "<p><b>" + tr("List subscription:") + "</b><br>https://lists.sourceforge.net/lists/listinfo/qmc2-devel</p>";
  labelProjectInfo->setText(projectInfoString);

  QString sysInfoString =
          "<p><b>" + tr("Build OS:") + "</b><br>" + XSTR(TARGET_OS_NAME) + " " + XSTR(TARGET_OS_RELEASE) + " " + XSTR(TARGET_MACHINE) + "</p>" +
          "<p><b>" + tr("Emulator version:") + "</b><br>" + qmc2Gamelist->mameVersion + "</p>" +
          "<p><b>" + tr("Template version:") + "</b><br>" + qmc2GlobalMAMEOptions->templateVersion + "</p>" +
          "<p><b>" + tr("Qt version:") + "</b><br>" + tr("Compile-time:") + " " + QT_VERSION_STR + "<br>" + tr("Run-time:") + " " + qVersion() + "</p>" +
          "<p><b>" + tr("Environment variables:") + "</b>";
  QStringList envVars = QProcess::systemEnvironment();
  envVars.sort();
  foreach (QString ev, envVars)
    sysInfoString += "<br>" + ev;
  sysInfoString += "</p>";
  textBrowserSystemInformation->setText(sysInfoString);

  // ensure a black background for the logo image
  labelLogoPixmap->setPalette(QPalette(QColor(0, 0, 0)));

  adjustSize();

  if ( widgetSize.isValid() ) {
    move(widgetPos);
    resize(widgetSize);
  }

  e->accept();
}

void About::hideEvent(QHideEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: About::hideEvent(QHideEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  widgetSize = size();
  widgetPos = pos();

  e->accept();
}
