#ifndef _MACROS_H_
#define _MACROS_H_

#include <unistd.h>

// max of two constants
#define MAX(a, b)		(((a) > (b)) ? (a) : (b))

// determine file existance
#define EXISTS(fn)		(access(fn, F_OK) == 0)

// determine file executable permission
#define IS_EXECUTABLE(fn)	(access(fn, X_OK) == 0)

// this is strange, but it's the only solution I found
// to make a string out of a non-string constant
#define STR(s)			#s
#define XSTR(s)			STR(s)

// index positions of game detail & image checker tabs
#define QMC2_PREVIEW_INDEX	0
#define QMC2_FLYER_INDEX	1
#define QMC2_ICON_INDEX		2
#define QMC2_MAMECONFIG_INDEX	2

// index positions of gamelist stack
#define QMC2_GAMELIST_INDEX	0
#define QMC2_HIERARCHY_INDEX	1

// search window timeout in milliseconds
#define QMC2_SEARCH_TIMEOUT	2000

// buffer size for reading zip-files
#define QMC2_ZIP_BUFFER_SIZE	32768

// buffer size for reading "normal" files
#define QMC2_FILE_BUFFER_SIZE	65536

// process timeouts in msecs
#define QMC2_PROCESS_POLL_TIME	25

// update lists every how many milliseconds (for image & sample checkers)?
#define QMC2_CHECK_UPDATE	1000

// maximium length of a single filename/path in a ZIP archive
#define QMC2_MAX_PATH_LENGTH	1024

// search delay in milliseconds (so it doesn't hamper typing)
#define QMC2_SEARCH_DELAY	250

// animation timeout in milliseconds (controls animation speeds)
#define QMC2_ANIMATION_TIMEOUT	250

// gamelist sort criteria
#define QMC2_SORT_BY_DESCRIPTION	0
#define QMC2_SORT_BY_ROM_STATE		1
#define QMC2_SORT_BY_YEAR		2
#define QMC2_SORT_BY_MANUFACTURER	3
#define QMC2_SORT_BY_NAME		4

// internal ROM state representations
#define QMC2_ROMSTATE_COUNT	5
#define QMC2_ROMSTATE_INT_C	0
#define QMC2_ROMSTATE_INT_M	1
#define QMC2_ROMSTATE_INT_U	2
#define QMC2_ROMSTATE_INT_I	3
#define QMC2_ROMSTATE_INT_N	4
#define QMC2_ROMSTATE_CHAR_C	'0'
#define QMC2_ROMSTATE_CHAR_M	'1'
#define QMC2_ROMSTATE_CHAR_U	'2'
#define QMC2_ROMSTATE_CHAR_I	'3'
#define QMC2_ROMSTATE_CHAR_N	'4'
#define QMC2_ROMSTATE_STRING_C	"0"
#define QMC2_ROMSTATE_STRING_M	"1"
#define QMC2_ROMSTATE_STRING_U	"2"
#define QMC2_ROMSTATE_STRING_I	"3"
#define QMC2_ROMSTATE_STRING_N	"4"

// ROM state filter responsiveness (update gamelist after how many milliseconds?)
#define QMC2_STATEFILTER_RESPONSIVENESS	1000

// responsiveness while reading the XML gamelist cache
#define QMC2_XMLCACHE_RESPONSIVENESS	50

// MAME option types
#define MAMEOPT_TYPE_UNKNOWN	0
#define MAMEOPT_TYPE_BOOL	1
#define MAMEOPT_TYPE_INT	2
#define MAMEOPT_TYPE_FLOAT	3
#define MAMEOPT_TYPE_STRING	4

// standard sizes
#define QMC2_ONE_KILOBYTE	1024
#define QMC2_ONE_MEGABYTE	1048576
#define QMC2_ONE_GIGABYTE	1073741824
#define QMC2_ONE_TERABYTE	1099511627776

// type conversions
#define QMC2_TO_UINT32(a)	((uchar)*((a) + 0) * (quint32)16777216ULL + \
                                 (uchar)*((a) + 1) * (quint32)65536ULL + \
                                 (uchar)*((a) + 2) * (quint32)256ULL + \
                                 (uchar)*((a) + 3))
#define QMC2_TO_UINT64(a)	((uchar)*((a) + 0) * (quint64)72057594037927936ULL + \
                                 (uchar)*((a) + 1) * (quint64)281474976710656ULL + \
                                 (uchar)*((a) + 2) * (quint64)1099511627776ULL + \
                                 (uchar)*((a) + 3) * (quint64)4294967296ULL + \
                                 (uchar)*((a) + 4) * (quint64)16777216ULL + \
                                 (uchar)*((a) + 5) * (quint64)65536ULL + \
                                 (uchar)*((a) + 6) * (quint64)256ULL + \
                                 (uchar)*((a) + 7))

#endif
