#include <QMessageBox>
#include <QPixmapCache>
#include <QHeaderView>
#include <QTextStream>
#include <QScrollBar>
#include <QProcess>
#include <QTimer>
#include <QTime>
#include <QFile>
#include <QMap>
#include <QStyleFactory>
#include <QBitArray>
#include <QFileInfo>
#include <QFileDialog>
#include <QAction>
#include <QPair>

#include "macros.h"
#include "qmc2main.h"
#include "options.h"
#include "mameopt.h"
#include "procmgr.h"
#include "gamelist.h"
#include "preview.h"
#include "flyer.h"
#include "dbrowser.h"
#include "about.h"
#include "welcome.h"
#include "imgcheck.h"
#include "sampcheck.h"
#include "romalyzer.h"
#include "unzip.h"
#if JOYSTICK == 1
#include "joystick.h"
#endif

// global variables
MainWindow *qmc2MainWindow = NULL;
Gamelist *qmc2Gamelist = NULL;
Options *qmc2Options = NULL;
QSettings *qmc2Config = NULL;
MameOptions *qmc2GlobalMAMEOptions = NULL;
MameOptions *qmc2MAMEOptions = NULL;
ProcessManager *qmc2ProcessManager = NULL;
Preview *qmc2Preview = NULL;
Flyer *qmc2Flyer = NULL;
About *qmc2About = NULL;
DocBrowser *qmc2DocBrowser = NULL;
Welcome *qmc2Welcome = NULL;
ImageChecker *qmc2ImageChecker = NULL;
SampleChecker *qmc2SampleChecker = NULL;
ROMAlyzer *qmc2ROMAlyzer = NULL;
bool qmc2ReloadActive = FALSE;
bool qmc2ImageCheckActive = FALSE;
bool qmc2SampleCheckActive = FALSE;
bool qmc2EarlyReloadActive = FALSE;
bool qmc2VerifyActive = FALSE;
bool qmc2FilterActive = FALSE;
bool qmc2ROMAlyzerActive = FALSE;
bool qmc2ROMAlyzerPaused = FALSE;
bool qmc2GuiReady = FALSE;
bool qmc2CleaningUp = FALSE;
bool qmc2StartingUp = TRUE;
bool qmc2EarlyStartup = TRUE;
bool qmc2ScaledPreview = TRUE;
bool qmc2ScaledFlyer = TRUE;
bool qmc2SmoothScaling = FALSE;
bool qmc2RetryLoadingImages = TRUE;
bool qmc2UsePreviewFile = FALSE;
bool qmc2UseFlyerFile = FALSE;
bool qmc2UseIconFile = FALSE;
bool qmc2IconsPreloaded = FALSE;
bool qmc2CheckItemVisibility = TRUE;
bool qmc2AutomaticReload = FALSE;
int qmc2GamelistResponsiveness = 100;
int qmc2UpdateDelay = 10;
QFile *qmc2FrontendLogFile = NULL;
QFile *qmc2MAMELogFile = NULL;
QTextStream qmc2FrontendLogStream;
QTextStream qmc2MAMELogStream;
QTranslator *qmc2Translator = NULL;
QTranslator *qmc2QtTranslator = NULL;
QString qmc2LastFrontendLogMessage = "";
quint64 qmc2FrontendLogMessageRepeatCount = 0;
QString qmc2LastMAMELogMessage = "";
quint64 qmc2MAMELogMessageRepeatCount = 0;
bool qmc2StopParser = FALSE;
QTreeWidgetItem *qmc2CurrentItem = NULL;
QTreeWidgetItem *qmc2LastItem = NULL;
QTreeWidgetItem *qmc2HierarchySelectedItem = NULL;
QMenu *qmc2EmulatorMenu = NULL,
      *qmc2GameMenu = NULL,
      *qmc2FavoritesMenu = NULL,
      *qmc2PlayedMenu = NULL,
      *qmc2SearchMenu = NULL;
QMap<QString, QTreeWidgetItem *> qmc2GamelistItemMap;
QMap<QString, QTreeWidgetItem *> qmc2HierarchyItemMap;
QMap<QString, QTreeWidgetItem *> qmc2GamelistItemByDescriptionMap;
QMap<QString, QString> qmc2GamelistNameMap;
QMap<QString, QString> qmc2GamelistDescriptionMap;
QMap<QString, QString> qmc2GamelistStatusMap;
QMap<QString, QStringList> qmc2HierarchyMap;
QMap<QString, QIcon> qmc2IconMap;
QMap<QString, QPair<QString, QAction *> > qmc2ShortcutMap;
QMap<QString, QString> qmc2CustomShortcutMap;
int qmc2SortCriteria = QMC2_SORT_BY_DESCRIPTION;
Qt::SortOrder qmc2SortOrder = Qt::AscendingOrder;
QBitArray qmc2Filter;
unzFile qmc2IconFile = NULL;
QStringList qmc2BiosROMs;
KeyPressFilter *qmc2KeyPressFilter = NULL;
QMap<QString, int> qmc2QtKeyMap;
#if JOYSTICK == 1
Joystick *qmc2Joystick = NULL;
#endif

void MainWindow::log(char logOrigin, QString message)
{
  if ( !qmc2GuiReady )
    return;

  QString timeString = QTime::currentTime().toString("hh:mm:ss.zzz");

  // count subsequent message duplicates
  switch ( logOrigin ) {
    case LOG_QMC2:
      if ( message == qmc2LastFrontendLogMessage ) {
        qmc2FrontendLogMessageRepeatCount++;
        return;
      } else {
        qmc2LastFrontendLogMessage = message;
        if ( qmc2FrontendLogMessageRepeatCount > 0 )
          message.prepend(tr("last message repeated %n time(s)\n", "", qmc2FrontendLogMessageRepeatCount) + timeString + ": ");
        qmc2FrontendLogMessageRepeatCount = 0;
      }
      break;

    case LOG_MAME:
      if ( message == qmc2LastMAMELogMessage ) {
        qmc2MAMELogMessageRepeatCount++;
        return;
      } else {
        qmc2LastMAMELogMessage = message;
        if ( qmc2MAMELogMessageRepeatCount > 0 )
          message.prepend(tr("last message repeated %n time(s)\n", "", qmc2MAMELogMessageRepeatCount) + timeString + ": ");
        qmc2MAMELogMessageRepeatCount = 0;
      }
      break;

    default:
      break;
  }

  QString msg = timeString + ": " + message;

  switch ( logOrigin ) {
    case LOG_QMC2:
      textBrowserFrontendLog->append(msg);
      textBrowserFrontendLog->horizontalScrollBar()->setValue(0);
      if ( !qmc2FrontendLogFile )
        if ( (qmc2FrontendLogFile = new QFile(qmc2Config->value("Frontend/FilesAndDirectories/LogFile", "data/log/qmc2.log").toString(), this)) == NULL )
          return;
      if ( qmc2FrontendLogFile->handle() == -1 )
        if ( qmc2FrontendLogFile->open(QIODevice::WriteOnly | QIODevice::Text) )
          qmc2FrontendLogStream.setDevice(qmc2FrontendLogFile);
        else
          return;
      qmc2FrontendLogStream << msg << "\n";
      qmc2FrontendLogStream.flush();
      break;

    case LOG_MAME:
      textBrowserMAMELog->append(msg);
      textBrowserMAMELog->horizontalScrollBar()->setValue(0);
      if ( !qmc2MAMELogFile )
        if ( (qmc2MAMELogFile = new QFile(qmc2Config->value("MAME/FilesAndDirectories/LogFile", "data/log/mame.log").toString(), this)) == NULL )
          return;
      if ( qmc2MAMELogFile->handle() == -1 )
        if ( qmc2MAMELogFile->open(QIODevice::WriteOnly | QIODevice::Text) )
          qmc2MAMELogStream.setDevice(qmc2MAMELogFile);
        else
          return;
      qmc2MAMELogStream << msg << "\n";
      qmc2MAMELogStream.flush();
      break;

    default:
      break;
  }
}

MainWindow::MainWindow(QWidget *parent)
  : QMainWindow(parent)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::MainWindow(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  setupUi(this);

  qmc2Gamelist = new Gamelist(this);

  statusBar()->setShown(qmc2Config->value("Frontend/GUI/Statusbar").toBool());

  // preview & flyer widgets
  QHBoxLayout *previewLayout = new QHBoxLayout, *flyerLayout = new QHBoxLayout;
  qmc2Preview = new Preview(tabPreview);
  previewLayout->addWidget(qmc2Preview);
  tabPreview->setLayout(previewLayout);
  qmc2Flyer = new Flyer(tabFlyer);
  flyerLayout->addWidget(qmc2Flyer);
  tabFlyer->setLayout(flyerLayout);

  // setup application style
  QString myStyle = qmc2Config->value("Frontend/GUI/Style", tr("Default")).toString();
  if ( myStyle != QObject::tr("Default") )
    setupStyle(myStyle);

  // restore layout
  if ( qmc2Config->value("Frontend/GUI/RestoreLayout").toBool() ) {
    log(LOG_QMC2, tr("restoring main widget layout"));
    move(qmc2Config->value("Frontend/Layout/MainWidget/Position", QPoint(0, 0)).toPoint());
    resize(qmc2Config->value("Frontend/Layout/MainWidget/Size", QSize(800, 600)).toSize());
    QList<int> hSplitterSizes, vSplitterSizes;
    QSize hSplitterSize = qmc2Config->value("Frontend/Layout/MainWidget/hSplitter").toSize();
    QSize vSplitterSize = qmc2Config->value("Frontend/Layout/MainWidget/vSplitter").toSize();
    if ( hSplitterSize.width() > 0 || hSplitterSize.height() > 0 )
      hSplitterSizes << hSplitterSize.width() << hSplitterSize.height();
    else
      hSplitterSizes << 100 << 100;
    if ( vSplitterSize.width() > 0 || vSplitterSize.height() > 0 )
      vSplitterSizes << vSplitterSize.width() << vSplitterSize.height();
    else
      vSplitterSizes << 100 << 100;
    hSplitter->setSizes(hSplitterSizes);
    vSplitter->setSizes(vSplitterSizes);
    tabWidgetGamelist->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/GamelistTab", 0).toInt());
    tabWidgetGameDetail->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/GameDetailTab", 0).toInt());
    tabWidgetLogsAndEmulators->setCurrentIndex(qmc2Config->value("Frontend/Layout/MainWidget/LogsAndEmulatorsTab", 0).toInt());
    int i;
    QVariantList defaultGamelistColumnWidths,
                 gamelistColumnWidths;
    defaultGamelistColumnWidths << 300 << 50 << 50 << 100;
    gamelistColumnWidths = qmc2Config->value("Frontend/Layout/MainWidget/GamelistColumnWidths", defaultGamelistColumnWidths).toList();
    for (i = 0; i < gamelistColumnWidths.count(); i++)
      treeWidgetGamelist->header()->resizeSection(i, gamelistColumnWidths[i].toInt());
    treeWidgetGamelist->header()->setDefaultAlignment(Qt::AlignLeft);
    QVariantList defaultHierarchyColumnWidths,
                 hierarchyColumnWidths;
    defaultHierarchyColumnWidths << 300 << 50 << 50 << 100;
    hierarchyColumnWidths = qmc2Config->value("Frontend/Layout/MainWidget/HierarchyColumnWidths", defaultHierarchyColumnWidths).toList();
    for (i = 0; i < hierarchyColumnWidths.count(); i++)
      treeWidgetHierarchy->header()->resizeSection(i, hierarchyColumnWidths[i].toInt());
    treeWidgetHierarchy->header()->setDefaultAlignment(Qt::AlignLeft);
    treeWidgetGamelist->header()->restoreState(qmc2Config->value("Frontend/Layout/MainWidget/GamelistHeaderState").toByteArray());
    treeWidgetHierarchy->header()->restoreState(qmc2Config->value("Frontend/Layout/MainWidget/HierarchyHeaderState").toByteArray());
    if ( qmc2Config->value("Frontend/Layout/ImageChecker/Visible").toBool() ) {
      on_actionCheckPreviews_activated();
      qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(qmc2Config->value("Frontend/ImageChecker/CurrentTab", 0).toInt());
    }
    if ( qmc2Config->value("Frontend/Layout/SampleChecker/Visible").toBool() )
      on_actionCheckSamples_activated();
    if ( qmc2Config->value("Frontend/Layout/ROMAlyzer/Visible").toBool() )
      on_actionROMAlyzer_activated();
  }

  // context menus
  QAction *action;
  QString s;

  qmc2EmulatorMenu = new QMenu(0);
  s = tr("Terminate selected emulator(s) (sends TERM signal to emulator process(es))");
  action = qmc2EmulatorMenu->addAction(tr("&Terminate"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/terminate.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_terminateEmulator_triggered()));
  s = tr("Kill selected emulator(s) (sends KILL signal to emulator process(es))");
  action = qmc2EmulatorMenu->addAction(tr("&Kill"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_killEmulator_triggered()));

  qmc2GameMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2GameMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2GameMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));
  qmc2GameMenu->addSeparator();
  s = tr("Check current game's ROM state");
  action = qmc2GameMenu->addAction(tr("Check &ROM state"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/rom.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionCheckCurrentROM_activated()));
  s = tr("Analyse current ROM with ROMAlyzer");
  action = qmc2GameMenu->addAction(tr("&Analyse ROM..."));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/search.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionAnalyseCurrentROM_activated()));

  qmc2SearchMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2SearchMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2SearchMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));
  qmc2SearchMenu->addSeparator();
  s = tr("Check current game's ROM state");
  action = qmc2SearchMenu->addAction(tr("Check &ROM state"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/rom.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionCheckCurrentROM_activated()));
  s = tr("Analyse current ROM with ROMAlyzer");
  action = qmc2SearchMenu->addAction(tr("&Analyse ROM..."));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/search.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionAnalyseCurrentROM_activated()));

  qmc2FavoritesMenu = new QMenu(0);
  s = tr("Play selected game");
  action = qmc2FavoritesMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  qmc2FavoritesMenu->addSeparator();
  s = tr("Check current game's ROM state");
  action = qmc2FavoritesMenu->addAction(tr("Check &ROM state"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/rom.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionCheckCurrentROM_activated()));
  s = tr("Analyse current ROM with ROMAlyzer");
  action = qmc2FavoritesMenu->addAction(tr("&Analyse ROM..."));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/search.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionAnalyseCurrentROM_activated()));
  qmc2FavoritesMenu->addSeparator();
  s = tr("Remove from favorites");
  action = qmc2FavoritesMenu->addAction(tr("&Remove"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/editdelete.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_removeFromFavorites_triggered()));
  s = tr("Clear all favorites");
  action = qmc2FavoritesMenu->addAction(tr("&Clear"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_clearAllFavorites_triggered()));
  s = tr("Save favorites now");
  action = qmc2FavoritesMenu->addAction(tr("&Save"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/filesave.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_saveFavorites_triggered()));

  qmc2PlayedMenu = new QMenu(0);
  action = qmc2PlayedMenu->addAction(tr("&Play"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/launch.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionPlay_activated()));
  s = tr("Add current game to favorites");
  action = qmc2PlayedMenu->addAction(tr("To &favorites"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/favorites.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionToFavorites_activated()));
  qmc2PlayedMenu->addSeparator();
  s = tr("Check current game's ROM state");
  action = qmc2PlayedMenu->addAction(tr("Check &ROM state"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/rom.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionCheckCurrentROM_activated()));
  s = tr("Analyse current ROM with ROMAlyzer");
  action = qmc2PlayedMenu->addAction(tr("&Analyse ROM..."));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/search.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(on_actionAnalyseCurrentROM_activated()));
  qmc2PlayedMenu->addSeparator();
  s = tr("Remove from played");
  action = qmc2PlayedMenu->addAction(tr("&Remove"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/editdelete.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_removeFromPlayed_triggered()));
  s = tr("Clear all played");
  action = qmc2PlayedMenu->addAction(tr("&Clear"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/kill.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_clearAllPlayed_triggered()));
  s = tr("Save play-history now");
  action = qmc2PlayedMenu->addAction(tr("&Save"));
  action->setToolTip(s); action->setStatusTip(s);
  action->setIcon(QIcon(QString::fromUtf8(":/data/img/filesave.png")));
  connect(action, SIGNAL(triggered()), this, SLOT(action_savePlayed_triggered()));

  // other actions
  connect(actionViewFullDetail, SIGNAL(triggered()), this, SLOT(viewFullDetail()));
  connect(actionViewParentClones, SIGNAL(triggered()), this, SLOT(viewParentClones()));
  connect(comboBoxViewSelect, SIGNAL(currentIndexChanged(int)), stackedWidgetView, SLOT(setCurrentIndex(int)));
  connect(&searchTimer, SIGNAL(timeout()), this, SLOT(on_comboBoxSearch_textChanged_delayed()));
  connect(&updateTimer, SIGNAL(timeout()), this, SLOT(on_treeWidgetGamelist_itemSelectionChanged_delayed()));

  treeWidgetGamelist->setFocus();
  comboBoxViewSelect->setCurrentIndex(qmc2Config->value("Frontend/GUI/GamelistView", QMC2_GAMELIST_INDEX).toInt());

  // restore toolbar state
  restoreState(qmc2Config->value("Frontend/Layout/MainWidget/ToolbarState", QByteArray()).toByteArray());

  QTimer::singleShot(0, this, SLOT(init()));
}

MainWindow::~MainWindow()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::~MainWindow()");
#endif
}

void MainWindow::on_actionPlay_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionPlay_activated()");
#endif

  if ( qmc2EarlyReloadActive )
    return;

  if ( !qmc2CurrentItem )
    return;

  QString gameName = qmc2CurrentItem->child(0)->text(1);

  if ( qmc2BiosROMs.contains(gameName) ) {
    // ROM is a BIOS and cannot be run...
    log(LOG_QMC2, tr("sorry, BIOS ROMs cannot be run")); 
    return;
  }

  qmc2LastItem = NULL;
  on_tabWidgetGameDetail_currentChanged(QMC2_MAMECONFIG_INDEX);

  QStringList args;
  QString sectionTitle;

  foreach ( sectionTitle, qmc2MAMEOptions->optionsMap.keys() ) {
    int i;
    for (i = 0; i < qmc2MAMEOptions->optionsMap[sectionTitle].count(); i++) {
      MameOption option = qmc2MAMEOptions->optionsMap[sectionTitle][i];
      QString globalOptionKey = "MAME/Configuration/Global/" + option.name;
      switch ( option.type ) {
        case MAMEOPT_TYPE_INT: {
          int  v = option.value.toInt();
          int dv = option.dvalue.toInt();
          int gv = qmc2Config->value(globalOptionKey, dv).toInt();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << QString("%1").arg(v);
          break;
        }

        case MAMEOPT_TYPE_FLOAT: {
          double  v = option.value.toDouble();
          double dv = option.dvalue.toDouble();
          double gv = qmc2Config->value(globalOptionKey, dv).toDouble();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << QString("%1").arg(v);
          break;
        }

        case MAMEOPT_TYPE_BOOL: {
          QString  v = option.value;
          QString dv = option.dvalue;
          QString gv = qmc2Config->value(globalOptionKey, dv).toString();
          if ( !option.valid )
            v = gv;
          if ( v != dv ) {
            if ( v == "true" )
              args << QString("-%1").arg(option.name);
            else
              args << QString("-no%1").arg(option.name);
          }
          break;
        }

        case MAMEOPT_TYPE_STRING:
        default: {
          QString  v = option.value;
          QString dv = option.dvalue;
          QString gv = qmc2Config->value(globalOptionKey, dv).toString();
          if ( !option.valid )
            v = gv;
          if ( v != dv )
            args << QString("-%1").arg(option.name) << v.replace("~", "$HOME");
          break;
        }
      }
    }
  }

  // start game
  args << gameName;
  QString command = qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString();
  qmc2ProcessManager->process(qmc2ProcessManager->start(command, args));

  // add game to played list
  listWidgetPlayed->blockSignals(TRUE);
  QList<QListWidgetItem *> matches = listWidgetPlayed->findItems(qmc2GamelistItemMap[gameName]->text(0), Qt::MatchExactly);
  QListWidgetItem *playedItem;
  if ( matches.count() > 0 ) {
    playedItem = listWidgetPlayed->takeItem(listWidgetPlayed->row(matches[0]));
  } else {
    playedItem = new QListWidgetItem();
    playedItem->setText(qmc2GamelistItemMap[gameName]->text(0));
  }
  listWidgetPlayed->insertItem(0, playedItem);
  listWidgetPlayed->setCurrentItem(playedItem);
  listWidgetPlayed->blockSignals(FALSE);
}

void MainWindow::on_actionToFavorites_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionToFavorites_activated()");
#endif

  if ( !qmc2CurrentItem )
    return;

  QString itemText = qmc2CurrentItem->text(0);

  if ( itemText == tr("Waiting for data...") )
    return;

  QList<QListWidgetItem *> matches = listWidgetFavorites->findItems(itemText, Qt::MatchExactly);
  if ( matches.count() <= 0 ) {
    QListWidgetItem *item = new QListWidgetItem(listWidgetFavorites);
    item->setText(itemText);
    listWidgetFavorites->sortItems();
  }
}

void MainWindow::on_actionReload_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionReload_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
    return;
  }

  if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("please wait for ROM verification to finish and try again"));
    return;
  }

  if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
    return;
  }

  if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
    return;
  }

  if ( qmc2ROMAlyzerActive ) {
    log(LOG_QMC2, tr("please wait for ROMAlyzer to finish the current analysis and try again"));
    return;
  }

  if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("gamelist reload already active"));
  } else {
    if ( qmc2Config->value("Frontend/GUI/SaveGameSelection").toBool() && !qmc2StartingUp ) {
      if ( qmc2CurrentItem ) {
        log(LOG_QMC2, tr("saving game selection"));
        qmc2Config->setValue("Frontend/Gamelist/SelectedGame", qmc2CurrentItem->text(0));
      } else
        qmc2Config->remove("Frontend/Gamelist/SelectedGame");
    }
    qmc2Gamelist->load();
  }
}

void MainWindow::on_actionExitStop_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionExitStop_activated()");
#endif

  close();
}

void MainWindow::on_actionCheckCurrentROM_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckCurrentROM_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
  } else if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("ROM verification already active"));
  } else if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
  } else if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
  } else if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
  } else if ( qmc2ROMAlyzerActive ) {
    log(LOG_QMC2, tr("please wait for ROMAlyzer to finish the current analysis and try again"));
  } else
    qmc2Gamelist->verify(TRUE);
}

void MainWindow::on_actionCheckROMs_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckROMs_activated()");
#endif

  if ( qmc2FilterActive ) {
    log(LOG_QMC2, tr("please wait for ROM state filter to finish and try again"));
  } else if ( qmc2VerifyActive ) {
    log(LOG_QMC2, tr("ROM verification already active"));
  } else if ( qmc2ReloadActive ) {
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
  } else if ( qmc2ImageCheckActive ) {
    log(LOG_QMC2, tr("please wait for image check to finish and try again"));
  } else if ( qmc2SampleCheckActive ) {
    log(LOG_QMC2, tr("please wait for sample check to finish and try again"));
  } else if ( qmc2ROMAlyzerActive ) {
    log(LOG_QMC2, tr("please wait for ROMAlyzer to finish the current analysis and try again"));
  } else {
    if ( !qmc2Gamelist->autoROMCheck ) {
      switch ( QMessageBox::question(this,
                                     tr("Confirm"),
                                     tr("The ROM verification process may be very time-consuming.\nIt will overwrite existing cached data.\n\nDo you really want to check all ROM states now?"),
                                     QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
          case QMessageBox::Yes:
            qmc2Gamelist->verify();
            break;

          default:
            break;
      }
    } else {
      if ( qmc2AutomaticReload )
        qmc2MainWindow->log(LOG_QMC2, tr("automatic ROM check triggered"));
      qmc2Gamelist->verify();
    }
  }
  qmc2Gamelist->autoROMCheck = FALSE;
  qmc2AutomaticReload = FALSE;
}

void MainWindow::on_actionCheckSamples_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckSamples_activated()");
#endif

  if ( !qmc2SampleChecker )
    qmc2SampleChecker = new SampleChecker(0);

  if ( qmc2SampleChecker->isHidden() )
    qmc2SampleChecker->show();
  else if ( qmc2SampleChecker->isMinimized() )
    qmc2SampleChecker->showNormal();

  QTimer::singleShot(0, qmc2SampleChecker, SLOT(raise()));
}

void MainWindow::on_actionCheckPreviews_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckPreviews_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_PREVIEW_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();

  QTimer::singleShot(0, qmc2ImageChecker, SLOT(raise()));
}

void MainWindow::on_actionCheckFlyers_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckFlyers_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_FLYER_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();

  QTimer::singleShot(0, qmc2ImageChecker, SLOT(raise()));
}

void MainWindow::on_actionCheckIcons_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionCheckIcons_activated()");
#endif

  if ( !qmc2ImageChecker )
    qmc2ImageChecker = new ImageChecker(0);

  qmc2ImageChecker->tabWidgetImageChecker->setCurrentIndex(QMC2_ICON_INDEX);
  if ( qmc2ImageChecker->isHidden() )
    qmc2ImageChecker->show();
  else if ( qmc2ImageChecker->isMinimized() )
    qmc2ImageChecker->showNormal();

  QTimer::singleShot(0, qmc2ImageChecker, SLOT(raise()));
}

void MainWindow::on_actionROMAlyzer_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionROMAlyzer_activated()");
#endif

  if ( !qmc2ROMAlyzer )
    qmc2ROMAlyzer = new ROMAlyzer(0);

  if ( !qmc2ROMAlyzerActive )
    qmc2ROMAlyzer->lineEditGames->setText("*");

  if ( qmc2ROMAlyzer->isHidden() )
    qmc2ROMAlyzer->show();
  else if ( qmc2ROMAlyzer->isMinimized() )
    qmc2ROMAlyzer->showNormal();

  QTimer::singleShot(0, qmc2ROMAlyzer, SLOT(raise()));
}

void MainWindow::on_actionAnalyseCurrentROM_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionAnalyseCurrentROM_activated()");
#endif

  if ( !qmc2CurrentItem )
    return;

  if ( qmc2ROMAlyzerActive ) {
    log(LOG_QMC2, tr("please wait for ROMAlyzer to finish the current analysis and try again"));
    return;
  }

  if ( !qmc2ROMAlyzer )
    qmc2ROMAlyzer = new ROMAlyzer(0);

  qmc2ROMAlyzer->lineEditGames->setText(qmc2CurrentItem->child(0)->text(1));

  if ( qmc2ROMAlyzer->isHidden() )
    qmc2ROMAlyzer->show();
  else if ( qmc2ROMAlyzer->isMinimized() )
    qmc2ROMAlyzer->showNormal();
  qmc2ROMAlyzer->raise();

  qApp->processEvents();
  QTimer::singleShot(0, qmc2ROMAlyzer->pushButtonAnalyze, SLOT(animateClick()));
}

void MainWindow::on_actionClearImageCache_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionClearImageCache_activated()");
#endif

  QPixmapCache::clear();
  log(LOG_QMC2, tr("image cache cleared"));
}

void MainWindow::on_actionClearIconCache_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionClearIconCache_activated()");
#endif

  qmc2IconMap.clear();
  qmc2IconsPreloaded = FALSE;
  log(LOG_QMC2, tr("icon cache cleared"));
}

void MainWindow::on_actionRecreateTemplateMap_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionRecreateTemplateMap_activated()");
#endif

  qmc2GlobalMAMEOptions->createTemplateMap();
}

void MainWindow::on_actionOptions_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionOptions_activated()");
#endif

  if ( qmc2Options->isHidden() )
    qmc2Options->show();
  else if ( qmc2Options->isMinimized() )
    qmc2Options->showNormal();

  QTimer::singleShot(0, qmc2Options, SLOT(raise()));
}

void MainWindow::on_actionDocumentation_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionDocumentation_activated()");
#endif

  if ( !qmc2DocBrowser ) {
    qmc2DocBrowser = new DocBrowser(this);
    QStringList searchPaths;
    searchPaths << qmc2Config->value("Frontend/FilesAndDirectories/DataDirectory").toString() +
                   "doc/html/" +
                   qmc2Config->value("Frontend/GUI/Language").toString();
    qmc2DocBrowser->manualBrowser->setSearchPaths(searchPaths);
    qmc2DocBrowser->manualBrowser->setSource(QUrl("index.html"));
  }

  qmc2DocBrowser->show();
  qmc2DocBrowser->raise();
}

void MainWindow::on_actionAbout_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionAbout_activated()");
#endif

  if ( !qmc2About )
    qmc2About = new About(this);

  qmc2About->show();
  qmc2About->raise();
}

void MainWindow::on_actionAboutQt_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionAboutQt_activated()");
#endif

  QMessageBox::aboutQt(this, tr("About Qt"));
}

void MainWindow::on_actionArcadeSetup_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionArcadeSetup_activated()");
#endif

}

void MainWindow::on_actionArcadeToggle_activated()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_actionArcadeToggle_activated()");
#endif

}

void MainWindow::on_comboBoxSearch_textChanged(QString)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_comboBoxSearch_textChanged(QString)");
#endif

  searchTimer.start(QMC2_SEARCH_DELAY);
}

void MainWindow::on_comboBoxSearch_textChanged_delayed()
{
  searchTimer.stop();

  QString pattern = comboBoxSearch->currentText();

  // easy pattern match
  pattern = "*" + pattern.replace(' ', "* *") + "*";
  pattern.replace(QString("*^"), "");
  pattern.replace(QString("$*"), "");

  listWidgetSearch->clear();
  QList<QTreeWidgetItem *> matches = treeWidgetGamelist->findItems(pattern, Qt::MatchContains | Qt::MatchWildcard);
  
  int i;
  for (i = 0; i < matches.count(); i++) {
    QListWidgetItem *item = new QListWidgetItem(listWidgetSearch);
    item->setText(matches.at(i)->text(0));
  }

  qmc2Gamelist->numSearchGames = matches.count();
  labelGamelistStatus->setText(qmc2Gamelist->status());

#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_comboBoxSearch_textChanged_delayed(QString pattern = \"" + pattern + "\"): matches.count() = " + QString::number(matches.count()));
#endif

  QTimer::singleShot(0, this, SLOT(checkCurrentSearchSelection()));
}

void MainWindow::on_comboBoxSearch_activated(QString pattern)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_comboBoxSearch_activated(QString pattern = \"" + pattern + "\")");
#endif

  on_comboBoxSearch_textChanged_delayed();
  if ( tabWidgetGamelist->currentWidget() != tabSearch ) {
    tabWidgetGamelist->setCurrentWidget(tabSearch);
    QTimer::singleShot(0, listWidgetSearch, SLOT(setFocus()));
  }
}

void MainWindow::on_listWidgetSearch_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_currentTextChanged(QString s = \""+ s + "\")");
#endif

  QTreeWidgetItem *matchItem = qmc2GamelistItemByDescriptionMap[s];
  if ( matchItem ) {
    qmc2CheckItemVisibility = FALSE;
    treeWidgetGamelist->clearSelection();
    treeWidgetGamelist->setCurrentItem(matchItem);
    qmc2CurrentItem = matchItem;
  }
#ifdef QMC2_DEBUG
  else
    log(LOG_QMC2, "DEBUG: ERROR: no match found (?)");
#endif
}

void MainWindow::on_listWidgetSearch_currentItemChanged(QListWidgetItem *current, QListWidgetItem *previous)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_listWidgetSearch_currentItemChanged(QListWidgetItem *current = %1, QListWidgetItem *previous = %2)").arg((qulonglong)current).arg((qulonglong)previous));
#endif

  QTreeWidgetItem *glItem = NULL;
  if ( current )
    glItem = qmc2GamelistItemByDescriptionMap[current->text()];
  if ( glItem ) {
    qmc2CheckItemVisibility = FALSE;
    treeWidgetGamelist->clearSelection();
    treeWidgetGamelist->setCurrentItem(glItem);
    qmc2CurrentItem = glItem;
  }
}

void MainWindow::on_listWidgetSearch_itemPressed(QListWidgetItem *current)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_listWidgetSearch_itemPressed(QListWidgetItem *current = %1)").arg((qulonglong)current));
#endif

  on_listWidgetSearch_currentItemChanged(current, NULL);
}

void MainWindow::on_listWidgetSearch_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_itemSelectionChanged()");
#endif

  QList<QListWidgetItem *> selected = listWidgetSearch->selectedItems();
  if ( selected.count() > 0 )
    on_listWidgetSearch_currentItemChanged(selected[0], NULL);
}

void MainWindow::on_listWidgetSearch_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  if ( item == NULL )
    return;

  QTreeWidgetItem *matchItem = qmc2GamelistItemByDescriptionMap[item->text()];
  if ( matchItem ) {
    qmc2CheckItemVisibility = FALSE;
    tabWidgetGamelist->setCurrentIndex(0);
    treeWidgetGamelist->clearSelection();
    treeWidgetGamelist->setCurrentItem(matchItem);
    treeWidgetGamelist->scrollToItem(matchItem, QAbstractItemView::PositionAtTop);
    qmc2CurrentItem = matchItem;
    if ( !qmc2ReloadActive )
      treeWidgetGamelist->expandItem(matchItem);
  }
  else
    log(LOG_QMC2, tr("ERROR: no match found (?)"));
}

void MainWindow::on_listWidgetFavorites_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_currentTextChanged(QString s = \""+ s + "\")");
#endif

  on_listWidgetSearch_currentTextChanged(s);
}

void MainWindow::on_listWidgetFavorites_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_itemSelectionChanged()");
#endif

  QList<QListWidgetItem *> selected = listWidgetFavorites->selectedItems();
  if ( selected.count() > 0 )
    on_listWidgetSearch_currentItemChanged(selected[0], NULL);
}

void MainWindow::on_listWidgetFavorites_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_listWidgetSearch_itemActivated(item);
}

void MainWindow::on_listWidgetPlayed_currentTextChanged(QString s)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_currentTextChanged(QString s = \""+ s + "\")");
#endif

  on_listWidgetSearch_currentTextChanged(s);
}

void MainWindow::on_listWidgetPlayed_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_itemSelectionChanged()");
#endif

  QList<QListWidgetItem *> selected = listWidgetPlayed->selectedItems();
  if ( selected.count() > 0 )
    on_listWidgetSearch_currentItemChanged(selected[0], NULL);
}

void MainWindow::on_listWidgetPlayed_itemActivated(QListWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_itemActivated(QListWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_listWidgetSearch_itemActivated(item);
}

void MainWindow::on_tabWidgetGamelist_currentChanged(int currentIndex)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_tabWidgetGamelist_currentChanged(int i = " + QString::number(currentIndex) + ")");
#endif

  switch ( currentIndex ) {
    case 0:
      QTimer::singleShot(0, this, SLOT(scrollToCurrentItem()));
      break;

    case 1:
      QTimer::singleShot(0, this, SLOT(checkCurrentSearchSelection()));
      break;

    case 2:
      QTimer::singleShot(0, this, SLOT(checkCurrentFavoritesSelection()));
      break;

    case 3:
      QTimer::singleShot(0, this, SLOT(checkCurrentPlayedSelection()));
      break;

    default:
      break;
  }
}

void MainWindow::scrollToCurrentItem()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::scrollToCurrentItem()");
#endif

  QTreeWidgetItem *ci = treeWidgetGamelist->currentItem();

  if ( qmc2CurrentItem )
    ci = qmc2CurrentItem;

  if ( ci ) {
    if ( ci->text(0) == tr("Waiting for data...") )
      return;

    switch ( stackedWidgetView->currentIndex() ) {
      case QMC2_HIERARCHY_INDEX:
        ci = qmc2HierarchyItemMap[ci->child(0)->text(1)];
        if ( ci ) {
          treeWidgetHierarchy->clearSelection();
          treeWidgetHierarchy->setCurrentItem(ci);
          treeWidgetHierarchy->scrollToItem(ci, QAbstractItemView::PositionAtTop);
        }
        break;

      case QMC2_GAMELIST_INDEX:
      default:
        qmc2CheckItemVisibility = FALSE;
        treeWidgetGamelist->clearSelection();
        treeWidgetGamelist->setCurrentItem(ci);
        treeWidgetGamelist->scrollToItem(ci, QAbstractItemView::PositionAtTop);
        break;
    }
    ci->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentSearchSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentSearchSelection()");
#endif

  listWidgetSearch->clearSelection();
  listWidgetSearch->setCurrentIndex(QModelIndex());

  if ( !qmc2CurrentItem )
    return;

  QList<QListWidgetItem *> searchMatches = listWidgetSearch->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( searchMatches.count() > 0 ) {
    listWidgetSearch->setCurrentItem(searchMatches[0]);
    listWidgetSearch->scrollToItem(searchMatches[0], QAbstractItemView::PositionAtTop);
    searchMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentFavoritesSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentFavoritesSelection()");
#endif

  listWidgetFavorites->clearSelection();
  listWidgetFavorites->setCurrentIndex(QModelIndex());

  if ( !qmc2CurrentItem )
    return;

  QList<QListWidgetItem *> favoritesMatches = listWidgetFavorites->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( favoritesMatches.count() > 0 ) {
    listWidgetFavorites->setCurrentItem(favoritesMatches[0]);
    listWidgetFavorites->scrollToItem(favoritesMatches[0], QAbstractItemView::PositionAtTop);
    favoritesMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::checkCurrentPlayedSelection()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::checkCurrentPlayedSelection()");
#endif

  listWidgetPlayed->clearSelection();
  listWidgetPlayed->setCurrentIndex(QModelIndex());

  if ( !qmc2CurrentItem )
    return;

  QList<QListWidgetItem *> playedMatches = listWidgetPlayed->findItems(qmc2CurrentItem->text(0), Qt::MatchExactly);
  if ( playedMatches.count() > 0 ) {
    listWidgetPlayed->setCurrentItem(playedMatches[0]);
    listWidgetPlayed->scrollToItem(playedMatches[0], QAbstractItemView::PositionAtTop);
    playedMatches[0]->setSelected(TRUE);
  }
}

void MainWindow::on_tabWidgetGameDetail_currentChanged(int currentIndex)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_tabWidgetGameDetail_currentChanged(int currentIndex = " + QString::number(currentIndex) + ")");
#endif

  if ( !qmc2CurrentItem )
    return;
 
  if ( qmc2EarlyReloadActive )
    return;

  if ( qmc2CurrentItem != qmc2LastItem ) {
    tabMAME->setUpdatesEnabled(FALSE);
    if ( qmc2MAMEOptions ) {
      qmc2MAMEOptions->save();
      QLayout *vbl = tabMAME->layout();
      delete vbl;
      delete qmc2MAMEOptions;
      delete pushButtonCurrentMameOptionsExportToFile;
      delete pushButtonCurrentMameOptionsImportFromFile;
      qmc2MAMEOptions = NULL;
    }

    if ( currentIndex == QMC2_MAMECONFIG_INDEX ) {
      QString gameName = qmc2CurrentItem->child(0)->text(1);
      log(LOG_QMC2, tr("processing MAME configuration for '%1'").arg(qmc2CurrentItem->text(0)));
      QVBoxLayout *layout = new QVBoxLayout;
      qmc2MAMEOptions = new MameOptions("MAME/Configuration/" + gameName, tabMAME);
      qmc2MAMEOptions->load();
      layout->addWidget(qmc2MAMEOptions);
#if defined(SDLMAME)
      QHBoxLayout *buttonLayout = new QHBoxLayout();
      pushButtonCurrentMameOptionsExportToFile = new QPushButton(tr("Export to..."), this);
      pushButtonCurrentMameOptionsExportToFile->setToolTip(QObject::tr("Export game-specific MAME configuration"));
      pushButtonCurrentMameOptionsExportToFile->setStatusTip(QObject::tr("Export game-specific MAME configuration"));
      pushButtonCurrentMameOptionsImportFromFile = new QPushButton(QObject::tr("Import from..."), this);
      pushButtonCurrentMameOptionsImportFromFile->setToolTip(QObject::tr("Import game-specific MAME configuration"));
      pushButtonCurrentMameOptionsImportFromFile->setStatusTip(QObject::tr("Import game-specific MAME configuration"));
      buttonLayout->addWidget(pushButtonCurrentMameOptionsExportToFile);
      buttonLayout->addWidget(pushButtonCurrentMameOptionsImportFromFile);
      layout->addLayout(buttonLayout);
      tabMAME->setLayout(layout);
      qmc2MAMEOptions->show();
      pushButtonCurrentMameOptionsExportToFile->show();
      pushButtonCurrentMameOptionsImportFromFile->show();
      // export/import menus
      qmc2MainWindow->selectMenuCurrentMameOptionsExportToFile = new QMenu(qmc2MainWindow->pushButtonCurrentMameOptionsExportToFile);
      connect(qmc2MainWindow->selectMenuCurrentMameOptionsExportToFile->addAction(QIcon(QString::fromUtf8(":/data/img/work.png")), tr("<inipath>/%1.ini").arg(gameName)), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonCurrentMameOptionsExportToFile_clicked()));
      connect(qmc2MainWindow->selectMenuCurrentMameOptionsExportToFile->addAction(QIcon(QString::fromUtf8(":/data/img/fileopen.png")), tr("Select file...")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonCurrentMameOptionsSelectExportFile_clicked()));
      qmc2MainWindow->pushButtonCurrentMameOptionsExportToFile->setMenu(qmc2MainWindow->selectMenuCurrentMameOptionsExportToFile);
      qmc2MainWindow->selectMenuCurrentMameOptionsImportFromFile = new QMenu(qmc2MainWindow->pushButtonCurrentMameOptionsImportFromFile);
      connect(qmc2MainWindow->selectMenuCurrentMameOptionsImportFromFile->addAction(QIcon(QString::fromUtf8(":/data/img/work.png")), tr("<inipath>/%1.ini").arg(gameName)), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonCurrentMameOptionsImportFromFile_clicked()));
      connect(qmc2MainWindow->selectMenuCurrentMameOptionsImportFromFile->addAction(QIcon(QString::fromUtf8(":/data/img/fileopen.png")), tr("Select file...")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonCurrentMameOptionsSelectImportFile_clicked()));
      qmc2MainWindow->pushButtonCurrentMameOptionsImportFromFile->setMenu(qmc2MainWindow->selectMenuCurrentMameOptionsImportFromFile);
#endif

      qmc2MAMEOptions->resizeColumnToContents(0);
      qmc2MAMEOptions->pseudoConstructor();
      qmc2LastItem = qmc2CurrentItem;
    }
    tabMAME->setUpdatesEnabled(TRUE);
  }
}

void MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_actionPlay_activated();
}

void MainWindow::on_treeWidgetHierarchy_itemActivated(QTreeWidgetItem *item)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemActivated(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  on_actionPlay_activated();
}

void MainWindow::on_treeWidgetGamelist_itemExpanded(QTreeWidgetItem *item)
{
  if ( qmc2ReloadActive ) {
    treeWidgetGamelist->collapseItem(item);
    log(LOG_QMC2, tr("please wait for reload to finish and try again"));
    return;
  }

#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemExpanded(QTreeWidgetItem *item = 0x"+ QString::number((ulong)item, 16) + ")");
#endif

  if ( item->child(0) ) {
    if ( item->child(0)->text(0) == tr("Waiting for data...") )
      qmc2Gamelist->parseGameDetail(item);
  }
}

void MainWindow::on_treeWidgetGamelist_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_currentItemChanged(QTreeWidgetItem *current = 0x" + QString::number((ulong)current, 16) + ", QTreeWidgetItem *previous = 0x" + QString::number((ulong)previous, 16) + ")");
#endif

  // workaround for a Qt bug: when POS1/Home is pressed, QTreeWidget & QTreeView don't correctly select the first VISIBLE item,
  // if the top item is HIDDEN
  if ( current ) {
    if ( qmc2CheckItemVisibility ) {
      if ( current->isHidden() ) {
        int i = treeWidgetGamelist->indexOfTopLevelItem(current);
        if ( i >= 0 ) {
          while ( current->isHidden() && i < treeWidgetGamelist->topLevelItemCount() ) {
            current = treeWidgetGamelist->topLevelItem(++i);
            if ( current == NULL ) 
              break;
          }
          if ( current ) {
            treeWidgetGamelist->setCurrentItem(current);
          }
        }
      }
    } else
      qmc2CurrentItem = current;
  }
  qmc2CheckItemVisibility = TRUE;
  on_tabWidgetGameDetail_currentChanged(tabWidgetGameDetail->currentIndex());
}

void MainWindow::on_treeWidgetHierarchy_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_currentItemChanged(QTreeWidgetItem *current = 0x" + QString::number((ulong)current, 16) + ", QTreeWidgetItem *previous = 0x" + QString::number((ulong)previous, 16) + ")");
#endif

  qmc2CheckItemVisibility = FALSE;
  on_tabWidgetGameDetail_currentChanged(tabWidgetGameDetail->currentIndex());
}

void MainWindow::on_treeWidgetGamelist_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemSelectionChanged()");
#endif

  updateTimer.start(qmc2UpdateDelay);
}

void MainWindow::on_treeWidgetGamelist_itemSelectionChanged_delayed()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_itemSelectionChanged_delayed()");
#endif

  updateTimer.stop();
  QList<QTreeWidgetItem *>selected;
  if ( qmc2HierarchySelectedItem != NULL )
    selected.append(qmc2HierarchySelectedItem);
  else
    selected = treeWidgetGamelist->selectedItems();
  if ( selected.count() == 0 )
    if ( treeWidgetGamelist->currentItem() )
      selected.append(treeWidgetGamelist->currentItem());
  if ( selected.count() > 0 ) {
    QTreeWidgetItem *topLevelItem = selected.at(0);
    while ( topLevelItem->parent() )
      topLevelItem = topLevelItem->parent();
    if ( topLevelItem ) {
      qmc2CurrentItem = topLevelItem;
      qmc2Preview->repaint();
      qmc2Flyer->repaint();
      on_tabWidgetGameDetail_currentChanged(tabWidgetGameDetail->currentIndex());
    }
  } else
    qmc2CurrentItem = NULL;
  qmc2HierarchySelectedItem = NULL;
}

void MainWindow::on_treeWidgetHierarchy_itemSelectionChanged()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_itemSelectionChanged()");
#endif

  qmc2HierarchySelectedItem = NULL;
  QList<QTreeWidgetItem *>selected = treeWidgetHierarchy->selectedItems();
  if ( selected.count() > 0 ) {
    if ( selected.at(0)->text(0) == tr("Waiting for data...") )
      return;
    treeWidgetGamelist->clearSelection();
    qmc2HierarchySelectedItem = qmc2GamelistItemByDescriptionMap[selected.at(0)->text(0)];
    qmc2CheckItemVisibility = FALSE;
    treeWidgetGamelist->setCurrentItem(qmc2HierarchySelectedItem);
  }
}

void MainWindow::on_textBrowserFrontendLog_textChanged()
{
  textBrowserFrontendLog->verticalScrollBar()->setSliderPosition(textBrowserFrontendLog->verticalScrollBar()->maximum());
  textBrowserFrontendLog->horizontalScrollBar()->setSliderPosition(textBrowserFrontendLog->horizontalScrollBar()->minimum());
}

void MainWindow::on_textBrowserMAMELog_textChanged()
{
  textBrowserMAMELog->verticalScrollBar()->setSliderPosition(textBrowserMAMELog->verticalScrollBar()->maximum());
  textBrowserMAMELog->horizontalScrollBar()->setSliderPosition(textBrowserMAMELog->horizontalScrollBar()->minimum());
}

void MainWindow::on_listWidgetEmulators_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetEmulators_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetEmulators->itemAt(p);
  if ( item ) {
    listWidgetEmulators->setItemSelected(item, TRUE);
    qmc2EmulatorMenu->move(listWidgetEmulators->viewport()->mapToGlobal(p));
    qmc2EmulatorMenu->show();
  }
}

void MainWindow::action_terminateEmulator_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_terminateEmulator_triggered()");
#endif

  QList<QListWidgetItem *> sl = listWidgetEmulators->selectedItems();
  int i;
  for (i = 0; i < sl.count(); i++) {
    QString s = sl.at(i)->text().remove("#");
    s.truncate(s.indexOf(":"));
    qmc2ProcessManager->terminate(s.toInt());
  }

  if ( sl.count() == 0 )
    if ( listWidgetEmulators->currentItem() ) {
      QString s = listWidgetEmulators->currentItem()->text().remove("#");
      s.truncate(s.indexOf(":"));
      qmc2ProcessManager->terminate(s.toInt());
    }
}

void MainWindow::action_killEmulator_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_killEmulator_triggered()");
#endif

  QList<QListWidgetItem *> sl = listWidgetEmulators->selectedItems();
  int i;
  for (i = 0; i < sl.count(); i++) {
    QString s = sl.at(i)->text().remove("#");
    s.truncate(s.indexOf(":"));
    qmc2ProcessManager->kill(s.toInt());
  }

  if ( sl.count() == 0 )
    if ( listWidgetEmulators->currentItem() ) {
      QString s = listWidgetEmulators->currentItem()->text().remove("#");
      s.truncate(s.indexOf(":"));
      qmc2ProcessManager->kill(s.toInt());
    }
}

void MainWindow::action_removeFromFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_removeFromFavorites_triggered()");
#endif

  QListWidgetItem *i = listWidgetFavorites->currentItem();
  if ( i ) {
    QListWidgetItem *item = listWidgetFavorites->takeItem(listWidgetFavorites->row(i));
    delete item;
  }
}

void MainWindow::action_clearAllFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_clearAllFavorites_triggered()");
#endif

  switch ( QMessageBox::question(this, tr("Confirm"),
                                 tr("Are you sure you want to clear the favorites list?"),
                                 QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
    case QMessageBox::No:
      return;
      break;

    case QMessageBox::Yes:
      listWidgetFavorites->clear();
      qmc2Gamelist->saveFavorites();
      break;

    default:
      break;
  }
}

void MainWindow::action_saveFavorites_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_saveFavorites_triggered()");
#endif

  qmc2Gamelist->saveFavorites();
}

void MainWindow::action_removeFromPlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_removeFromPlayed_triggered()");
#endif

  QListWidgetItem *i = listWidgetPlayed->currentItem();
  if ( i ) {
    QListWidgetItem *item = listWidgetPlayed->takeItem(listWidgetPlayed->row(i));
    delete item;
  }
}

void MainWindow::action_clearAllPlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_clearAllPlayed_triggered()");
#endif

  switch ( QMessageBox::question(this, tr("Confirm"),
                                 tr("Are you sure you want to clear the play history?"),
                                 QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
    case QMessageBox::No:
      return;
      break;

    case QMessageBox::Yes:
      listWidgetPlayed->clear();
      qmc2Gamelist->savePlayHistory();
      break;

    default:
      break;
  }
}

void MainWindow::action_savePlayed_triggered()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::action_savePlayed_triggered()");
#endif

  qmc2Gamelist->savePlayHistory();
}

void MainWindow::on_listWidgetSearch_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetSearch_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetSearch->itemAt(p);
  if ( item ) {
    listWidgetSearch->setItemSelected(item, TRUE);
    qmc2SearchMenu->move(listWidgetSearch->viewport()->mapToGlobal(p));
    qmc2SearchMenu->show();
  }
}

void MainWindow::on_listWidgetFavorites_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetFavorites_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetFavorites->itemAt(p);
  if ( item ) {
    listWidgetFavorites->setItemSelected(item, TRUE);
    qmc2FavoritesMenu->move(listWidgetFavorites->viewport()->mapToGlobal(p));
    qmc2FavoritesMenu->show();
  }
}

void MainWindow::on_listWidgetPlayed_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_listWidgetPlayed_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QListWidgetItem *item = listWidgetPlayed->itemAt(p);
  if ( item ) {
    listWidgetPlayed->setItemSelected(item, TRUE);
    qmc2PlayedMenu->move(listWidgetPlayed->viewport()->mapToGlobal(p));
    qmc2PlayedMenu->show();
  }
}

void MainWindow::on_treeWidgetGamelist_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetGamelist_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QTreeWidgetItem *item = treeWidgetGamelist->itemAt(p);
  if ( !item )
    return;
  if ( item->text(0) == tr("Waiting for data...") )
    return;
  if ( item ) {
    treeWidgetGamelist->setItemSelected(item, TRUE);
    qmc2GameMenu->move(treeWidgetGamelist->viewport()->mapToGlobal(p));
    qmc2GameMenu->show();
  }
}

void MainWindow::on_treeWidgetHierarchy_customContextMenuRequested(const QPoint &p)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_treeWidgetHierarchy_customContextMenuRequested(const QPoint &p = ...)");
#endif

  QTreeWidgetItem *item = treeWidgetHierarchy->itemAt(p);
  if ( !item )
    return;
  if ( item->text(0) == tr("Waiting for data...") )
    return;
  if ( item ) {
    treeWidgetHierarchy->setItemSelected(item, TRUE);
    qmc2GameMenu->move(treeWidgetHierarchy->viewport()->mapToGlobal(p));
    qmc2GameMenu->show();
  }
}

void MainWindow::on_stackedWidgetView_currentChanged(int index)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_tackedWidgetView_currentChanged(int index = %1)").arg(index));
#endif

  if ( !qmc2CurrentItem )
    return;

  if ( qmc2CurrentItem->text(0) == tr("Waiting for data...") )
    return;

  QTreeWidgetItem *hierarchyItem;
  switch ( index ) {
    case QMC2_HIERARCHY_INDEX:
      hierarchyItem = qmc2HierarchyItemMap[qmc2CurrentItem->child(0)->text(1)];
      treeWidgetHierarchy->clearSelection();
      if ( hierarchyItem ) {
        treeWidgetHierarchy->setCurrentItem(hierarchyItem);
        treeWidgetHierarchy->scrollToItem(hierarchyItem, QAbstractItemView::PositionAtTop);
        hierarchyItem->setSelected(TRUE);
      }
      break;

    case QMC2_GAMELIST_INDEX:
    default:
      scrollToCurrentItem();
      break;
  }
}

void MainWindow::on_pushButtonGlobalMameOptionsSelectExportFile_clicked()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_pushButtonGlobalMameOptionsSelectExportFile_clicked()");
#endif

#if defined(SDLMAME)
  QString fileName = qmc2Config->value("MAME/Configuration/Global/inipath", getenv("HOME")).toString().replace("~", getenv("HOME")) + "/mame.ini";
#else
  QString fileName;
#endif
  QString s = QFileDialog::getSaveFileName(qmc2Options, tr("Choose export file"), fileName, tr("All files (*)"));
  if ( !s.isNull() )
    on_pushButtonGlobalMameOptionsExportToFile_clicked(s);
}

void MainWindow::on_pushButtonGlobalMameOptionsExportToFile_clicked(QString useFileName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_pushButtonGlobalMameOptionsExportToFile_clicked(QString useFileName = %1)").arg(useFileName));
#endif

  qmc2GlobalMAMEOptions->exportToIni(TRUE, useFileName);
}

void MainWindow::on_pushButtonGlobalMameOptionsSelectImportFile_clicked()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_pushButtonGlobalMameOptionsSelectImportFile_clicked()");
#endif

#if defined(SDLMAME)
  QString fileName = qmc2Config->value("MAME/Configuration/Global/inipath", getenv("HOME")).toString().replace("~", getenv("HOME")) + "/mame.ini";
#else
  QString fileName;
#endif
  QString s = QFileDialog::getOpenFileName(qmc2Options, tr("Choose import file"), fileName, tr("All files (*)"));
  if ( !s.isNull() )
    on_pushButtonGlobalMameOptionsImportFromFile_clicked(s);
}

void MainWindow::on_pushButtonGlobalMameOptionsImportFromFile_clicked(QString useFileName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_pushButtonGlobalMameOptionsImportFromFile_clicked(QString useFileName = %1)").arg(useFileName));
#endif

  qmc2GlobalMAMEOptions->importFromIni(TRUE, useFileName);
}

void MainWindow::on_pushButtonCurrentMameOptionsSelectExportFile_clicked()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_pushButtonCurrentMameOptionsSelectExportFile_clicked()");
#endif

  QStringList iniPaths = qmc2Config->value("MAME/Configuration/Global/inipath", getenv("HOME")).toString().split(";");
  QString iniPath;
  if ( iniPaths.count() > 0 ) {
    iniPath = iniPaths[0].replace("~", getenv("HOME"));
  } else {
    iniPath = ".";
    log(LOG_QMC2, tr("WARNING: invalid inipath"));
  }
  QString s = QFileDialog::getSaveFileName(this, tr("Choose export file"), iniPath, tr("All files (*)"));
  if ( !s.isNull() )
    on_pushButtonCurrentMameOptionsExportToFile_clicked(s);
}

void MainWindow::on_pushButtonCurrentMameOptionsExportToFile_clicked(QString useFileName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_pushButtonCurrentMameOptionsExportToFile_clicked(QString useFileName = %1)").arg(useFileName));
#endif

  qmc2MAMEOptions->exportToIni(FALSE, useFileName);
}

void MainWindow::on_pushButtonCurrentMameOptionsSelectImportFile_clicked()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::on_pushButtonCurrentMameOptionsSelectImportFile_clicked()");
#endif

  if ( !qmc2CurrentItem )
    return;

  QStringList iniPaths = qmc2Config->value("MAME/Configuration/Global/inipath", getenv("HOME")).toString().split(";");
  QString iniPath;
  if ( iniPaths.count() > 0 ) {
    iniPath = iniPaths[0].replace("~", getenv("HOME"));
  } else {
    iniPath = ".";
    log(LOG_QMC2, tr("WARNING: invalid inipath"));
  }
  iniPath += "/" + qmc2CurrentItem->child(0)->text(1) + ".ini";
  QString s = QFileDialog::getOpenFileName(this, tr("Choose import file"), iniPath, tr("All files (*)"));
  if ( !s.isNull() )
    on_pushButtonCurrentMameOptionsImportFromFile_clicked(s);
}

void MainWindow::on_pushButtonCurrentMameOptionsImportFromFile_clicked(QString useFileName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_pushButtonCurrentMameOptionsImportFromFile_clicked(QString useFileName = %1)").arg(useFileName));
#endif

  qmc2MAMEOptions->importFromIni(FALSE, useFileName);
}

#if JOYSTICK == 1
#define QMC2_DEBUG
void MainWindow::on_joystickAxisValueChanged(int axis, int value)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_joystickAxisValueChanged(int axis = %1, int value = %2)").arg(axis).arg(value));
#endif

}

void MainWindow::on_joystickButtonValueChanged(int button, bool value)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_joystickButtonValueChanged(int button = %1, bool value = %2)").arg(button).arg(value));
#endif

}

void MainWindow::on_joystickHatValueChanged(int hat, int value)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_joystickHatValueChanged(int hat = %1, int value = %2)").arg(hat).arg(value));
#endif

}

void MainWindow::on_joystickTrackballValueChanged(int trackball, int deltaX, int deltaY)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::on_joystickTrackballValueChanged(int trackball = %1, int deltaX = %2, int deltaY = %3)").arg(trackball).arg(deltaX).arg(deltaY));
#endif

}
#undef QMC2_DEBUG
#endif

void MainWindow::closeEvent(QCloseEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::closeEvent(QCloseEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( qmc2CleaningUp ) {
    e->ignore();
    return;
  }

  if ( qmc2ReloadActive || qmc2VerifyActive || qmc2FilterActive || qmc2ImageCheckActive || qmc2SampleCheckActive || qmc2ROMAlyzerActive ) {
    qmc2StopParser = TRUE;
    log(LOG_QMC2, tr("stopping current processing upon user request"));
    e->ignore();
    return;
  }

  if ( !qmc2Options->applied ) {
    switch ( QMessageBox::question(this, tr("Confirm"), tr("Your configuration changes have not been applied yet.\nReally quit?"), QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton) ) {
      case QMessageBox::No:
        e->ignore();
        return;
        break;

      default:
        break;
    }
  }
 
  bool doKillEmulators = qmc2Config->value("Frontend/GUI/KillEmulatorsOnExit").toBool();
  if ( qmc2ProcessManager->procMap.count() > 0 && !doKillEmulators ) {
    switch ( QMessageBox::question(this, tr("Confirm"), tr("There are one or more emulators still running.\nShould they be killed on exit?"), QMessageBox::Yes, QMessageBox::No, QMessageBox::Cancel) ) {
      case QMessageBox::Yes:
        doKillEmulators = TRUE;
        break;

      case QMessageBox::No:
        doKillEmulators = FALSE;
        break;

      case QMessageBox::Cancel:
        e->ignore();
        return;
        break;

      default:
        break;
    }
  }

  qmc2CleaningUp = TRUE;
  log(LOG_QMC2, tr("cleaning up"));

  if ( listWidgetFavorites->count() > 0 )
    qmc2Gamelist->saveFavorites();

  if ( listWidgetPlayed->count() > 0 )
    qmc2Gamelist->savePlayHistory();

  qmc2Config->setValue("Frontend/GUI/GamelistView", comboBoxViewSelect->currentIndex());
  if ( qmc2Config->value("Frontend/GUI/SaveGameSelection").toBool() ) {
    if ( qmc2CurrentItem ) {
      log(LOG_QMC2, tr("saving game selection"));
      qmc2Config->setValue("Frontend/Gamelist/SelectedGame", qmc2CurrentItem->text(0));
    } else
      qmc2Config->remove("Frontend/Gamelist/SelectedGame");
  }

  if ( qmc2Config->value("Frontend/GUI/SaveLayout").toBool() ) {
    log(LOG_QMC2, tr("saving main widget layout"));
    qmc2Config->setValue("Frontend/Layout/MainWidget/Position", pos());
    qmc2Config->setValue("Frontend/Layout/MainWidget/Size", size());
    qmc2Config->setValue("Frontend/Layout/MainWidget/hSplitter", QSize(hSplitter->sizes().at(0), hSplitter->sizes().at(1)));
    qmc2Config->setValue("Frontend/Layout/MainWidget/vSplitter", QSize(vSplitter->sizes().at(0), vSplitter->sizes().at(1)));
    qmc2Config->setValue("Frontend/Layout/MainWidget/GamelistTab", tabWidgetGamelist->currentIndex());
    qmc2Config->setValue("Frontend/Layout/MainWidget/GameDetailTab", tabWidgetGameDetail->currentIndex());
    qmc2Config->setValue("Frontend/Layout/MainWidget/LogsAndEmulatorsTab", tabWidgetLogsAndEmulators->currentIndex());
    // save toolbar state
    qmc2Config->setValue("Frontend/Layout/MainWidget/ToolbarState", saveState());
    int i;
    QVariantList gamelistColumnWidths;
    for (i = 0; i < treeWidgetGamelist->header()->count(); i++)
      gamelistColumnWidths << treeWidgetGamelist->header()->sectionSize(i);
    qmc2Config->setValue("Frontend/Layout/MainWidget/GamelistColumnWidths", gamelistColumnWidths);
    QVariantList hierarchyColumnWidths;
    for (i = 0; i < treeWidgetHierarchy->header()->count(); i++)
      hierarchyColumnWidths << treeWidgetHierarchy->header()->sectionSize(i);
    qmc2Config->setValue("Frontend/Layout/MainWidget/HierarchyColumnWidths", hierarchyColumnWidths);
    qmc2Config->setValue("Frontend/Layout/MainWidget/GamelistHeaderState", treeWidgetGamelist->header()->saveState());
    qmc2Config->setValue("Frontend/Layout/MainWidget/HierarchyHeaderState", treeWidgetHierarchy->header()->saveState());
  }

  if ( qmc2MAMEOptions ) {
    log(LOG_QMC2, tr("destroying current game's emulator configuration"));
    qmc2MAMEOptions->save();
    delete qmc2MAMEOptions;
  }
  log(LOG_QMC2, tr("destroying global emulator options"));
  //delete qmc2GlobalMAMEOptions; <- doing so will end up in heavy CPU load and a close-timeout at exit for Qt 4.3+
  //                              <- qmc2GlobalMAMEOptions->setParent(0) fixes this (this is strange but true :)
  qmc2GlobalMAMEOptions->pseudoDestructor();
  qmc2GlobalMAMEOptions->setParent(0);
  log(LOG_QMC2, tr("destroying gamelist"));
  delete qmc2Gamelist;
  log(LOG_QMC2, tr("destroying preview"));
  delete qmc2Preview;
  log(LOG_QMC2, tr("destroying flyer"));
  delete qmc2Flyer;
  if ( qmc2About ) {
    log(LOG_QMC2, tr("destroying about dialog"));
    delete qmc2About;
  }
  if ( qmc2DocBrowser ) {
    log(LOG_QMC2, tr("destroying documentation browser"));
    delete qmc2DocBrowser;
  }
  if ( qmc2ImageChecker ) {
    log(LOG_QMC2, tr("destroying image checker"));
    qmc2ImageChecker->close();
    delete qmc2ImageChecker;
  }
  if ( qmc2SampleChecker ) {
    log(LOG_QMC2, tr("destroying sample checker"));
    qmc2SampleChecker->close();
    delete qmc2SampleChecker;
  }
  if ( qmc2ROMAlyzer ) {
    log(LOG_QMC2, tr("destroying ROMAlyzer"));
    qmc2ROMAlyzer->saveState();
    delete qmc2ROMAlyzer;
  }

  log(LOG_QMC2, tr("destroying process manager"));
  if ( qmc2ProcessManager->procMap.count() > 0 ) {
    if ( doKillEmulators ) {
      log(LOG_QMC2, tr("killing %n running emulator(s) on exit", "", qmc2ProcessManager->procMap.count()));
      delete qmc2ProcessManager;
    } else
      log(LOG_QMC2, tr("keeping %n running emulator(s) alive", "", qmc2ProcessManager->procMap.count()));
  } else 
    delete qmc2ProcessManager;

  log(LOG_QMC2, tr("so long and thanks for all the fish"));

  delete qmc2KeyPressFilter;
  delete qmc2Options;
  e->accept();
}

void MainWindow::showEvent(QShowEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::showEvent(QShowEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  e->accept();
}

void MainWindow::hideEvent(QHideEvent *e)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::hideEvent(QHideEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  e->accept();
}

void MainWindow::init()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::init()");
#endif

  treeWidgetGamelist->header()->setClickable(FALSE);
  treeWidgetHierarchy->header()->setClickable(FALSE);
  on_actionReload_activated();
  qmc2GhostImagePixmap.load(":/data/img/ghost.png");

  qmc2EarlyStartup = FALSE;
}

void MainWindow::setupStyle(QString styleName)
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, QString("DEBUG: MainWindow::setupStyle(QString &styleName = %1").arg(styleName));
#endif

  static QPalette customPalette = QApplication::palette();

  QApplication::setStyle(QStyleFactory::create(styleName));
  if ( qmc2Config->value("Frontend/GUI/StandardColorPalette").toBool() )
    QApplication::setPalette(QApplication::style()->standardPalette());
  else
    QApplication::setPalette(customPalette);
}

void MainWindow::viewFullDetail()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::viewFullDetail()");
#endif

  comboBoxViewSelect->setCurrentIndex(QMC2_GAMELIST_INDEX);
  qApp->processEvents();
  tabWidgetGamelist->setCurrentIndex(QMC2_GAMELIST_INDEX);
  treeWidgetGamelist->setFocus();
}

void MainWindow::viewParentClones()
{
#ifdef QMC2_DEBUG
  log(LOG_QMC2, "DEBUG: MainWindow::viewParentClones()");
#endif

  comboBoxViewSelect->setCurrentIndex(QMC2_HIERARCHY_INDEX);
  qApp->processEvents();
  tabWidgetGamelist->setCurrentIndex(QMC2_GAMELIST_INDEX);
  treeWidgetHierarchy->setFocus();
}

bool KeyPressFilter::eventFilter(QObject *object, QEvent *event)
{
  if (event->type() == QEvent::KeyPress) {
    QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);

#ifdef QMC2_DEBUG
    qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: KeyPressFilter::eventFilter(QObject *object = %1, QEvent *event = %2)").arg((qulonglong)object).arg((qulonglong)event));
#endif
    
    if ( keyEvent->text() == "QMC2_EMULATED_KEY" ) {
#ifdef QMC2_DEBUG
      qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: emulated key event"));
#endif
      return FALSE;
    }

    QString pressedKey = QKeySequence(keyEvent->key()).toString();
    QString matchedKey = qmc2CustomShortcutMap.key(pressedKey);
    if ( !matchedKey.isEmpty() ) {
      if ( matchedKey != pressedKey ) {
#ifdef QMC2_DEBUG
        qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: key '%1' pressed, emulating key event for '%2'").arg(pressedKey).arg(matchedKey));
#endif
        // emulate a key event for the mapped key
        QKeyEvent *emulatedKeyEvent = new QKeyEvent(QEvent::KeyPress, qmc2QtKeyMap[matchedKey], Qt::NoModifier, "QMC2_EMULATED_KEY");
        qApp->postEvent(object, emulatedKeyEvent);
        return TRUE;
      } else {
#ifdef QMC2_DEBUG
        qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: key '%1' pressed, default key event processing").arg(pressedKey));
#endif
        // default key event processing
        return FALSE;
      }
    }

    QMap<QString, QString>::const_iterator it = qmc2CustomShortcutMap.find(pressedKey);
    if ( it != qmc2CustomShortcutMap.end() ) {
      if ( !it.value().isEmpty() ) {
#ifdef QMC2_DEBUG
        qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: key '%1' pressed, key event suppressed").arg(pressedKey));
#endif
        return TRUE;
      } else {
#ifdef QMC2_DEBUG
        qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: key '%1' pressed, default key event processing").arg(pressedKey));
#endif
        return FALSE;
      }
    }

#ifdef QMC2_DEBUG
    qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: key '%1' pressed, default key event processing").arg(pressedKey));
#endif
    return FALSE;
  } else {
    // default event processing
    return QObject::eventFilter(object, event);
  }
}

void myQtMessageHandler(QtMsgType type, const char *msg)
{
  if ( !qmc2GuiReady )
    return;

  switch ( type ) {
    case QtDebugMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtDebugMsg: " + QString(msg));
      break;

    case QtWarningMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtWarningMsg: " + QString(msg));
      break;

    case QtCriticalMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtCriticalMsg: " + QString(msg));
      break;

    case QtFatalMsg:
      qmc2MainWindow->log(LOG_QMC2, "QtFatalMsg: " + QString(msg));
      break;

    default:
      break;
  }
}

void prepareShortcuts()
{
  // shortcuts
  qmc2ShortcutMap["Ctrl+1"].second = qmc2MainWindow->actionCheckROMs;
  qmc2ShortcutMap["Ctrl+2"].second = qmc2MainWindow->actionCheckSamples;
  qmc2ShortcutMap["Ctrl+3"].second = qmc2MainWindow->actionCheckPreviews;
  qmc2ShortcutMap["Ctrl+4"].second = qmc2MainWindow->actionCheckFlyers;
  qmc2ShortcutMap["Ctrl+5"].second = qmc2MainWindow->actionCheckIcons;
  qmc2ShortcutMap["Ctrl+A"].second = qmc2MainWindow->actionAbout;
  qmc2ShortcutMap["Ctrl+D"].second = qmc2MainWindow->actionAnalyseCurrentROM;
  qmc2ShortcutMap["Ctrl+F"].second = qmc2MainWindow->actionToFavorites;
  qmc2ShortcutMap["Ctrl+H"].second = qmc2MainWindow->actionDocumentation;
  qmc2ShortcutMap["Ctrl+I"].second = qmc2MainWindow->actionClearImageCache;
  qmc2ShortcutMap["Ctrl+L"].second = qmc2MainWindow->actionArcadeSetup;
  qmc2ShortcutMap["Ctrl+N"].second = qmc2MainWindow->actionClearIconCache;
  qmc2ShortcutMap["Ctrl+O"].second = qmc2MainWindow->actionOptions;
  qmc2ShortcutMap["Ctrl+P"].second = qmc2MainWindow->actionPlay;
  qmc2ShortcutMap["Ctrl+Q"].second = qmc2MainWindow->actionAboutQt;
  qmc2ShortcutMap["Ctrl+R"].second = qmc2MainWindow->actionReload;
  qmc2ShortcutMap["Ctrl+S"].second = qmc2MainWindow->actionCheckCurrentROM;
  qmc2ShortcutMap["Ctrl+T"].second = qmc2MainWindow->actionRecreateTemplateMap;
  qmc2ShortcutMap["Ctrl+X"].second = qmc2MainWindow->actionExitStop;
  qmc2ShortcutMap["Ctrl+Z"].second = qmc2MainWindow->actionROMAlyzer;
  qmc2ShortcutMap["F5"].second = qmc2MainWindow->actionViewFullDetail;
  qmc2ShortcutMap["F6"].second = qmc2MainWindow->actionViewParentClones;
  qmc2ShortcutMap["F11"].second = qmc2MainWindow->actionArcadeToggle;

  // special keys
  qmc2ShortcutMap["+"].second = NULL;
  qmc2QtKeyMap["+"] = Qt::Key_Plus;
  qmc2ShortcutMap["-"].second = NULL;
  qmc2QtKeyMap["-"] = Qt::Key_Minus;
  qmc2ShortcutMap["Down"].second = NULL;
  qmc2QtKeyMap["Down"] = Qt::Key_Down;
  qmc2ShortcutMap["End"].second = NULL;
  qmc2QtKeyMap["End"] = Qt::Key_End;
  qmc2ShortcutMap["Esc"].second = NULL;
  qmc2QtKeyMap["Esc"] = Qt::Key_Escape;
  qmc2ShortcutMap["Left"].second = NULL;
  qmc2QtKeyMap["Left"] = Qt::Key_Left;
  qmc2ShortcutMap["Home"].second = NULL;
  qmc2QtKeyMap["Home"] = Qt::Key_Home;
  qmc2ShortcutMap["PgDown"].second = NULL;
  qmc2QtKeyMap["PgDown"] = Qt::Key_PageDown;
  qmc2ShortcutMap["PgUp"].second = NULL;
  qmc2QtKeyMap["PgUp"] = Qt::Key_PageUp;
  qmc2ShortcutMap["Return"].second = NULL;
  qmc2QtKeyMap["Return"] = Qt::Key_Return;
  qmc2ShortcutMap["Right"].second = NULL;
  qmc2QtKeyMap["Right"] = Qt::Key_Right;
  qmc2ShortcutMap["Tab"].second = NULL;
  qmc2QtKeyMap["Tab"] = Qt::Key_Tab;
  qmc2ShortcutMap["Up"].second = NULL;
  qmc2QtKeyMap["Up"] = Qt::Key_Up;

  qmc2Options->setupShortcutActions();
}

int main(int argc, char *argv[])
{
  qInstallMsgHandler(myQtMessageHandler);
  QApplication qmc2App(argc, argv);
  Q_INIT_RESOURCE(qmc2);
  qmc2App.setWindowIcon(QIcon(QString::fromUtf8(":/data/img/mame.png")));

  int checkReturn = QDialog::Accepted;
  qmc2Welcome = new Welcome(0);
  if ( !qmc2Welcome->checkOkay )
    checkReturn = qmc2Welcome->exec();
  delete qmc2Welcome;
  if ( checkReturn != QDialog::Accepted )
    exit(1);

  // setup key event filter
  qmc2KeyPressFilter = new KeyPressFilter(0);
  qmc2App.installEventFilter(qmc2KeyPressFilter);

  qmc2Options = new Options(0);
  qmc2Config = qmc2Options->config;
  qmc2ProcessManager = new ProcessManager(0);
  qmc2MainWindow = new MainWindow(0);

  qmc2Options->apply();
  qmc2MainWindow->show();
  qmc2MainWindow->raise();

  prepareShortcuts();

  qmc2GuiReady = TRUE;

  // greeting
  QString greeting = QObject::tr("M.A.M.E. Catalog / Launcher II v") +
                     QString::number(MAJOR) + "." + QString::number(MINOR) +
#ifdef BETA
                     ".b" + QString::number(BETA) +
#endif
                     " (Qt " + qVersion() +
#if defined(SDLMAME)
                     ", SDLMAME, " +
#else
                     ", ???, " +
#endif
                     qmc2Config->value("Frontend/GUI/Language").toString() + ")";
  qmc2MainWindow->log(LOG_QMC2, greeting);

  qmc2MainWindow->log(LOG_QMC2, QObject::tr("processing global emulator configuration"));
  QVBoxLayout *layout = new QVBoxLayout;
  qmc2GlobalMAMEOptions = new MameOptions("MAME/Configuration/Global", qmc2Options->tabGlobalConfiguration);
  qmc2GlobalMAMEOptions->load();
#if defined(SDLMAME)
  layout->addWidget(qmc2GlobalMAMEOptions);
  QHBoxLayout *buttonLayout = new QHBoxLayout();
  qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile = new QPushButton(QObject::tr("Export to..."), qmc2Options);
  qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile->setToolTip(QObject::tr("Export global MAME configuration"));
  qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile->setToolTip(QObject::tr("Export global MAME configuration"));
  qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile = new QPushButton(QObject::tr("Import from..."), qmc2Options);
  qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile->setToolTip(QObject::tr("Import global MAME configuration"));
  qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile->setStatusTip(QObject::tr("Import global MAME configuration"));
  buttonLayout->addWidget(qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile);
  buttonLayout->addWidget(qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile);
  layout->addLayout(buttonLayout);
  qmc2Options->tabGlobalConfiguration->setLayout(layout);
  qmc2GlobalMAMEOptions->show();
  qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile->show();
  qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile->show();
  // export/import menus
  qmc2MainWindow->selectMenuGlobalMameOptionsExportToFile = new QMenu(qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile);
  QObject::connect(qmc2MainWindow->selectMenuGlobalMameOptionsExportToFile->addAction(QIcon(QString::fromUtf8(":/data/img/work.png")), QObject::tr("<inipath>/mame.ini")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonGlobalMameOptionsExportToFile_clicked()));
  QObject::connect(qmc2MainWindow->selectMenuGlobalMameOptionsExportToFile->addAction(QIcon(QString::fromUtf8(":/data/img/fileopen.png")), QObject::tr("Select file...")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonGlobalMameOptionsSelectExportFile_clicked()));
  qmc2MainWindow->pushButtonGlobalMameOptionsExportToFile->setMenu(qmc2MainWindow->selectMenuGlobalMameOptionsExportToFile);
  qmc2MainWindow->selectMenuGlobalMameOptionsImportFromFile = new QMenu(qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile);
  QObject::connect(qmc2MainWindow->selectMenuGlobalMameOptionsImportFromFile->addAction(QIcon(QString::fromUtf8(":/data/img/work.png")), QObject::tr("<inipath>/mame.ini")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonGlobalMameOptionsImportFromFile_clicked()));
  QObject::connect(qmc2MainWindow->selectMenuGlobalMameOptionsImportFromFile->addAction(QIcon(QString::fromUtf8(":/data/img/fileopen.png")), QObject::tr("Select file...")), SIGNAL(triggered()), qmc2MainWindow, SLOT(on_pushButtonGlobalMameOptionsSelectImportFile_clicked()));
  qmc2MainWindow->pushButtonGlobalMameOptionsImportFromFile->setMenu(qmc2MainWindow->selectMenuGlobalMameOptionsImportFromFile);
#endif
  qmc2GlobalMAMEOptions->pseudoConstructor();

  return qmc2App.exec();
}
